/*
* This file is part of Bielefeld Type Library (BTL).
*
* Copyright(c) Frederic Siepmann
* http://opensource.cit-ec.de/projects/btl
*
* This file may be licensed under the terms of of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*
* Contributors: Leon Ziegler, Johannes Wienke, Matthias Schröder, Norman Köster,
* 				Lukas Twardon, Torben Töniges, Annika Rothert, Marian Pohling,
* 				Denis Schulze, Viktor Richter, Birte Carlmeyer, Maikel Linke
*/

/*
 * PerceptList.h
 *
 *  Created on: Aug 18, 2009
 *      Author: jwienke
 */

#ifndef PERCEPTLIST_H_
#define PERCEPTLIST_H_

#include <vector>

#include <boost/shared_ptr.hpp>

#include "Type.h"

namespace airobots {

/**
 * A special deleter for boost smart pointers that does not delete the
 * contents of the data because they are already managed by another smart
 * pointer.
 */
template<class P>
struct smart_pointer_deleter {
private:
	P p_;

public:

	inline smart_pointer_deleter(P const & p) :
		p_(p) {
	}

	inline void operator()(void const *) {
		p_.reset();
	}

	inline P const & get() const {
		return p_;
	}
};

class BadSharedPtrCast: public std::exception {
public:

	BadSharedPtrCast(std::string message) :
		message(message) {
	}

	const char* what() const throw () {
		return message.c_str();
	}

	virtual ~BadSharedPtrCast() throw () {

	}

private:
	std::string message;

};

/**
 * Makes a boost::shared_ptr with a different type from another shared_ptr.
 * Therefore dynamic_cast is used.
 *
 * @throws BadSharedPtrCast if types cannot be converted
 */
template<class A, class X>
inline static boost::shared_ptr<A> make_shared_from_another(
		boost::shared_ptr<X> qx) {
	A* newContents = dynamic_cast<A*> (qx.get());
	if (!newContents) {
		std::stringstream msg;
		msg << "Unable to convert type " << typeid(qx.get()).name()
				<< " to type " << typeid(newContents).name() << ".";
		throw BadSharedPtrCast(msg.str());
	}
	boost::shared_ptr<A> px(newContents, smart_pointer_deleter<
			boost::shared_ptr<X> > (qx));
	return px;
}

/**
 * A generic list of percepts. Percepts can be all classes that derive from
 * <code>Percept</code> and provide xmltio convert and extract methods.
 *
 * Internally all operations are done using boost::shared_ptr. This enables
 * real inheritance with re-interpretation of the contents using dynamic_cast.
 *
 * Attention: This class cannot be extracted using the typical xmltio extract
 * methods. Instead <code>xmltio::extract_list</code> must be used!
 *
 * @author jwienke
 */
template<typename T>
class PerceptList: public Type {

public:

	static std::string getBaseTag() {
		return "PERCEPTS";
	}

	static std::string getTemplate() {
		return std::string("<") + getBaseTag() + " type=\"\">" + "<GENERATOR/>"
				+ Timestamp::getTemplate() + "</" + getBaseTag() + ">";
	}

	PerceptList() {
	}

	/**
	 * Conversion constructor that enables something like real inheritance with
	 * sub-classes of <code>Percept</code>.
	 *
	 * @throws BadSharedPtrCast if types cannot be converted
	 */
	template<typename F>
	PerceptList(PerceptList<F> list) :
		Type(list.getGenerator(), list.getTimestamp()), percepts(
				list.getPerceptsAsType<T> ()) {
	}

	virtual ~PerceptList() {

	}

	/**
	 * Adds a percept to this list.
	 */
	void addPercept(boost::shared_ptr<T> percept) {
		percepts.push_back(percept);
	}

	/** Removes all percepts from list. */
	void clearList() {
		percepts.clear();
	}

	size_t size() const {
		return percepts.size();
	}

	/**
	 * Adds a percept to this list and therefore creates a new smart pointer
	 * with newly allocated memory.
	 */
	void addPercept(T percept) {
		percepts.push_back(boost::shared_ptr<T>(new T(percept)));
	}

	const std::vector<boost::shared_ptr<T> > & getPercepts() const {
		return percepts;
	}

	/**
	 * Returns the contents of this percept list as another type that can be
	 * achieved via dynamic_cast.
	 *
	 * @throws BadSharedPtrCast if T cannot be casted to F
	 */
	template<typename F>
	std::vector<boost::shared_ptr<F> > getPerceptsAsType() const {

		std::vector<boost::shared_ptr<F> > vect;
		typename std::vector<boost::shared_ptr<T> >::const_iterator it =
				percepts.begin();
		for (; it != percepts.end(); ++it) {
			boost::shared_ptr<F> f = make_shared_from_another<F, T> (*it);
			vect.push_back(f);
		}
		return vect;

	}

	/**
	 * Returns the name of the percepts contained in this list.
	 */
	std::string getPerceptName() {
		return T::getPerceptName();
	}

	void sanitizeLocation(xmltio::Location &loc) const {
		// not used here because of template specialization below
	}

	void fillInto(xmltio::Location &loc) const {
		// not needed here because of template specialization of convert below
	}

	static void fromLocation(const xmltio::Location &loc, PerceptList &type) {
		// not needed here because of sepcial extract method below
	}

private:
	std::vector<boost::shared_ptr<T> > percepts;

};

}

namespace xmltio {

template<typename T>
void convert(const airobots::PerceptList<T> & list, Location & loc) {	

	// sanitize location
	xmltio::XPath xp(T::getBaseTag());
	while (XPathIterator i = xp.evaluate(loc)) {
		Location l(*i);
		loc.remove(l);
	}	

	loc[XPath("GENERATOR")] = list.getGenerator();
	loc[XPath(airobots::Timestamp::getBaseTag())] = list.getTimestamp();

	loc["type"] = T::getPerceptName();

	typename std::vector<boost::shared_ptr<T> >::const_iterator it;
	for (it = list.getPercepts().begin(); it != list.getPercepts().end(); ++it) {
		Location perceptLoc(T::getTemplate(), T::getBaseTag());
		perceptLoc = *(it->get());
		loc.add(perceptLoc);
	}

}

template<typename T>
airobots::PerceptList<T> extract_list(const Location & loc) {

	airobots::PerceptList<T> list;

	airobots::Type::fromLocation(loc, list);

	for (XPathIterator i = XPath(T::getBaseTag()).evaluate(loc); i; ++i) {
		Location l(*i);

		T rawT = extract<T> (l);
		list.addPercept(boost::shared_ptr<T>(new T(rawT)));

	}

	return list;

}

}

#endif /* PERCEPTLIST_H_ */
