/*
* This file is part of Bielefeld Type Library (BTL).
*
* Copyright(c) Frederic Siepmann
* http://opensource.cit-ec.de/projects/btl
*
* This file may be licensed under the terms of of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*
* Contributors: Leon Ziegler, Johannes Wienke, Matthias Schröder, Norman Köster,
* 				Lukas Twardon, Torben Töniges, Annika Rothert, Marian Pohling,
* 				Denis Schulze, Viktor Richter, Birte Carlmeyer, Maikel Linke
*/

/*
 * Rotation3D.h
 *
 *  Created on: Apr 11, 2011
 *      Author: lziegler
 */

#ifndef ROTATION3D_H_
#define ROTATION3D_H_

#include <xmltio/xmltio.hpp>
#include <Eigen/Dense>
#include <Eigen/Geometry>
#include "Type.h"
#include "UnitConverter.h"
#include "Point3D.h"

namespace airobots {
/**
 * This class represents an arbitrary rotation in 3D space. Several conversions
 * are supported. For theoretical background see
 * http://www.euclideanspace.com/maths/geometry/rotations/
 *
 * @author lziegler
 */
class Rotation3D: public Type {
public:

	/**
	 * Create rotation object representing no rotation (identity).
	 */
	Rotation3D();

	/**
	 * Create rotation object from axis/angle representation.
	 *
	 * @param x
	 *            First component of the axis.
	 * @param y
	 *            Second component of the axis.
	 * @param z
	 *            Third component of the axis.
	 * @param angle
	 *            Rotation angle around the defined axis.
	 * @param angleUnit
	 *            Unit of the defined angle.
	 */
	Rotation3D(double x, double y, double z, double angle, AngleUnit angleUnit);

	/**
	 * Create rotation object from axis/angle representation.
	 *
	 * @param axis
	 *            The axis as vecmath object.
	 * @param angle
	 *            Rotation angle around the defined axis.
	 * @param angleUnit
	 *            Unit of the defined angle.
	 * @see #Rotation3D(double, double, double, double, AngleUnit)
	 */
	Rotation3D(const Eigen::Vector3d &axis, double angle, AngleUnit angleUnit);

	/**
	 * Create rotation object from quaternion representation.
	 *
	 * @param quaternion
	 *            The quaternion object.
	 */
	Rotation3D(const Eigen::Quaterniond &quaternion);

	/**
	 * Create rotation object from euler angles.
	 * The conversion uses conventions as described on page:
	 *  http://www.euclideanspace.com/maths/geometry/rotations/euler/index.htm
	 *
	 *   Coordinate System: right hand
	 *   Positive angle: right hand
	 *   Order of euler angles: heading first, then attitude, then bank
	 *
	 * @param heading
	 *            first rotation (azimuth, yaw, theta)
	 * @param attitude
	 *            second rotation (elevation, pitch, phi)
	 * @param bank
	 *            third rotation (tilt, psi, roll)
	 * @param angleUnit
	 *            Unit of the rotations
	 */
	Rotation3D(double heading, double attitude, double bank,
			AngleUnit angleUnit);

	/**
	 * Create rotation object from euler angles.
	 * The conversion uses conventions as described on page:
	 *  http://www.euclideanspace.com/maths/geometry/rotations/euler/index.htm
	 *
	 *   Coordinate System: right hand
	 *   Positive angle: right hand
	 *   Order of euler angles: heading first, then attitude, then bank
	 *
	 * @param vec
	 *            A vector containing the angles for heading, attitude and bank.
	 * @param angleUnit
	 *            Unit of the rotations
	 * @see #Rotation3D(double, double, double, AngleUnit)
	 */
	Rotation3D(const Eigen::Vector3d &vec, AngleUnit angleUnit);

	/**
	 * Create rotation object from rotation matrix.
	 *
	 * @param mat
	 *            The rotation matrix.
	 */
	Rotation3D(const Eigen::Matrix3d &mat);

	/**
	 * Destructor.
	 */
	virtual ~Rotation3D();

	/**
	 * Set rotation from matrix.
	 *
	 * @param mat
	 *            The rotation matrix.
	 */
	void setMatrix(const Eigen::Matrix3d &mat);

	/**
	 * Set rotation from axis/angle representation.
	 *
	 * @param x
	 *            First component of the axis.
	 * @param y
	 *            Second component of the axis.
	 * @param z
	 *            Third component of the axis.
	 * @param angle
	 *            Rotation angle around the defined axis.
	 * @param angleUnit
	 *            Unit of the defined angle.
	 */
	void setAxisAngle(double x, double y, double z, double angle,
			AngleUnit angleUnit);

	/**
	 * Set rotation from axis/angle representation.
	 *
	 * @param axis
	 *            The axis as a vector object.
	 * @param angle
	 *            Rotation angle around the defined axis.
	 * @param angleUnit
	 *            Unit of the defined angle.
	 * @see #Rotation3D(double, double, double, double, AngleUnit)
	 */
	void setAxisAngle(const Eigen::Vector3d &axis, double angle, AngleUnit angleUnit);

	/**
	 * Set rotation from axis representation.
	 *
	 * @param a
	 *	     The axis as an AngleAxisd object
	 */
	
	void setAxisAngle(const Eigen::AngleAxisd &a);
	
	/**
	 * Set rotation from quaternion representation.
	 *
	 * @param rotation
	 *            The rotation.
	 */
	
	void setQuaternion(const Eigen::Quaterniond &quaternion);

	/**
	 * Set rotation from euler angles.
	 * The conversion uses conventions as described on page:
	 *  http://www.euclideanspace.com/maths/geometry/rotations/euler/index.htm
	 *
	 *   Coordinate System: right hand
	 *   Positive angle: right hand
	 *   Order of euler angles: heading first, then attitude, then bank
	 *
	 * @param heading
	 *            first rotation (azimuth, yaw, theta)
	 * @param attitude
	 *            second rotation (elevation, pitch, phi)
	 * @param bank
	 *            third rotation (tilt, psi, roll)
	 * @param angleUnit
	 *            Unit of the rotations
	 */
	void setEuler(double heading, double attitude, double bank,
			AngleUnit angleUnit);

	/**
	 * Set rotation from euler angles.
	 * The conversion uses conventions as described on page:
	 *  http://www.euclideanspace.com/maths/geometry/rotations/euler/index.htm
	 *
	 *   Coordinate System: right hand
	 *   Positive angle: right hand
	 *   Order of euler angles: heading first, then attitude, then bank
	 *
	 * @param vec
	 *            A vector containing the angles for heading, attitude and bank.
	 * @param angleUnit
	 *            Unit of the rotations
	 * @see #setEuler(double, double, double, AngleUnit)
	 */
	void setEuler(const Eigen::Vector3d &vec, AngleUnit angleUnit);

	/**
	 * Returns the rotation in matrix representation.
	 *
	 * @return The rotation matrix.
	 */
	Eigen::Matrix3d getMatrix() const;

	/**
	 * Returns the rotation in axis/angle representation.
	 *
	 * @param axis
	 *            Reference to the target axis object (output argument).
	 * @param angle
	 *            Reference to the target angle (output argument).
	 * @param angleUnit
	 *            Unit of the angle value.
	 */
	void getAxisAngle(Eigen::Vector3d &axis, double &angle, AngleUnit angleUnit) const;

	/**
	 * Returns the rotation in quaternion representation.
	 *
	 * @return A quaternion object containing the rotation values.
	 */
	Eigen::Quaterniond getQuaternion() const;

	/**
	 * Returns the rotation in euler angles representation.
	 * The conversion uses conventions as described on page:
	 *  http://www.euclideanspace.com/maths/geometry/rotations/euler/index.htm
	 *
	 *   Coordinate System: right hand
	 *   Positive angle: right hand
	 *   Order of euler angles: heading first, then attitude, then bank
	 *
	 * @param unit
	 *            The desired unit of the angles.
	 * @return A vector object containing the rotation angles.
	 */
	Eigen::Vector3d getEuler(AngleUnit angleUnit) const;

	static std::string getBaseTag();
	static std::string getTemplate();

	bool operator==(const Rotation3D &other) const;

	void fillInto(xmltio::Location &loc) const;
	static void fromLocation(const xmltio::Location &loc, Rotation3D &type);
	void sanitizeLocation(xmltio::Location &loc) const;

private:
	xmltio::Location createValueLocation(double value) const;
	Eigen::Matrix3d mat;
};

}

#endif /* ROTATION3D_H_ */
