/*
 * This file is part of Bielefeld Type Library (BTL).
 * 
 * Copyright(c) Frederic Siepmann http://opensource.cit-ec.de/projects/btl
 * 
 * This file may be licensed under the terms of of the GNU Lesser General Public
 * License Version 3 (the ``LGPL''), or (at your option) any later version.
 * 
 * Software distributed under the License is distributed on an ``AS IS'' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the LGPL for the
 * specific language governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this program. If not,
 * go to http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 * USA.
 * 
 * The development of this software was supported by the Excellence Cluster EXC
 * 277 Cognitive Interaction Technology. The Excellence Cluster EXC 277 is a
 * grant of the Deutsche Forschungsgemeinschaft (DFG) in the context of the
 * German Excellence Initiative.
 * 
 * Contributors: Leon Ziegler, Johannes Wienke, Matthias Schröder, Norman
 * Köster, Lukas Twardon, Torben Töniges, Annika Rothert, Marian Pohling, Denis
 * Schulze, Viktor Richter, Birte Carlmeyer, Maikel Linke
 */
package de.unibi.airobots.btl.data;

import java.lang.reflect.Method;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.RejectedExecutionException;

import nu.xom.Attribute;
import nu.xom.Element;
import nu.xom.Node;
import nu.xom.Nodes;

/**
 * Results of the object recognition. This is a very general base class.
 * 
 * @author sebschne
 * @author jwienke
 */
public class ObjectData extends Type {

	protected final static String CLASS_ELEMENT = "CLASS";
	protected final static String RATING_ELEMENT = "RATING";
	protected final static String RELIABILITY_ATTRIBUTE = "RELIABILITY";

	public static class Hypothesis extends Type {

		/**
		 * Reliability of the recognition result.
		 */
		private double reliability;

		public void setReliability(double reliability) {
			this.reliability = reliability;
		}

		/**
		 * Returns a calculated reliability for this object.
		 * 
		 * @return reliability value. Range: 0...1
		 */
		public double getReliability() {
			return reliability;
		}

		/**
		 * Getter for the xml base tag used for this (de-)serialization.
		 * 
		 * @return xml base tag
		 */
		public static String getBaseTag() {
			return "HYPOTHESIS";
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public void fillInto(Element parent) {

			super.fillInto(parent);

			// reliability
			Element rating = new Element(RATING_ELEMENT);
			Element reliability = new Element(RELIABILITY_ATTRIBUTE);
			reliability.addAttribute(new Attribute("value", String
					.valueOf(getReliability())));
			rating.appendChild(reliability);

			parent.appendChild(rating);
		}

		/**
		 * Fills an {@link ObjectData.Hypothesis} object from a given XOM
		 * {@link Element}.
		 * 
		 * @param hypothesisElement
		 *            The XOM {@link Element} to fill an object from.
		 * @param type
		 *            The {@link ObjectData.Hypothesis} object to fill with all
		 *            the information given by the {@link Element} object.
		 */
		public static void fromElement(Element hypothesisElement,
				Hypothesis type) {

			Type.fromElement(hypothesisElement, type);

			Element rating = hypothesisElement
					.getFirstChildElement(RATING_ELEMENT);

			String relString = rating
					.getFirstChildElement(RELIABILITY_ATTRIBUTE)
					.getAttribute("value").getValue();
			if (relString != null && !relString.equals("")) {
				type.reliability = Double.parseDouble(relString);
			} else {
				type.reliability = 0d;
			}
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public String toString() {
			return "[HYPOTHESIS timestamp: " + getTimestamp()
					+ " reliability:  " + getReliability() + "]";

		}
	}

	/**
	 * All hypotheses for this object.
	 */
	Set<Hypothesis> hypotheses = new HashSet<Hypothesis>();

	/**
	 * Name of this object.
	 */
	private String classLabel;

	/**
	 * Default constructor.
	 */
	public ObjectData() {
		setLegacyParsing(true);
	}

	/**
	 * Getter for the xml base tag used for this (de-)serialization.
	 * 
	 * @return xml base tag
	 */
	public static String getBaseTag() {
		return "OBJECT";
	}

	/**
	 * Fills an {@link ObjectData} object from a given XOM {@link Element}.
	 * 
	 * @param objectElement
	 *            The XOM {@link Element} to fill an object from.
	 * @param type
	 *            The {@link ObjectData} object to fill with all the information
	 *            given by the {@link Element} object.
	 */
	public static void fromElement(Element objectElement, ObjectData type) {

		Type.fromElement(objectElement, type);

		try {

			// check hypotheses
			Nodes nodes = objectElement.query(Hypothesis.getBaseTag());
			for (int i = 0; i < nodes.size(); i++) {
				Node node = nodes.get(i);
				if (node instanceof Element) {

					Hypothesis hyp = new Hypothesis();
					Hypothesis.fromElement((Element) node, hyp);
					type.addHypothesis(hyp);
				}
			}

			Element classes = objectElement.getFirstChildElement(CLASS_ELEMENT);

			final String classLabel = classes.getValue();
			if (classLabel == null) {
				throw new IllegalArgumentException("Empty class text.");
			}
			type.classLabel = classLabel;

		} catch (NullPointerException ex) {
			// this happens when an element or attribute that is required is
			// not present
			throw new IllegalArgumentException("Missing element or attribute "
					+ "in document.", ex);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toString() {
		String s = "[OBJECTDATA class: " + getClassLabel();
		for (Hypothesis hyp : hypotheses) {
			s += "\n\t" + hyp;
		}
		s += "\n]";

		return s;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void sanitizeElement(Element parent) {
		super.sanitizeElement(parent);

		Nodes nodes = parent.query(CLASS_ELEMENT);
		for (int i = 0; i < nodes.size(); i++) {
			parent.removeChild(nodes.get(i));
		}

		nodes = parent.query(Hypothesis.getBaseTag());
		for (int i = 0; i < nodes.size(); i++) {
			parent.removeChild(nodes.get(i));
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void fillInto(Element parent) {

		sanitizeElement(parent);

		super.fillInto(parent);

		for (Hypothesis hyp : hypotheses) {
			Element hypothesis = new Element(Hypothesis.getBaseTag());
			hyp.fillInto(hypothesis);
			parent.appendChild(hypothesis);
		}

		// class label
		Element className = new Element(CLASS_ELEMENT);
		className.appendChild(getClassLabel());

		parent.appendChild(className);
	}

	/**
	 * Returns the name of the recognized object.
	 * 
	 * @return
	 */
	public String getClassLabel() {
		return classLabel;
	}

	public Set<Hypothesis> getHypotheses() {
		return hypotheses;
	}

	public void addHypothesis(Hypothesis hypothesis) {
		this.hypotheses.add(hypothesis);
	}

	public void clearHypotheses() {
		hypotheses.clear();
	}

	/**
	 * Helper method to make sure that a given method was overwritten at runtime
	 * by any subclass.
	 * 
	 * @param superClass
	 *            Class that tests the given method. Usually the class invoking
	 *            this method.
	 * @param methodName
	 *            The name of the method to test.
	 * @param parameterTypes
	 *            Parameter types of the method to test.
	 * @throws RejectedExecutionException
	 *             Is thrown if the given method was not overwritten.
	 */
	protected void assureOverride(Class<?> superClass, String methodName,
			Class<?>... parameterTypes) throws RejectedExecutionException {

		if (!(this.getClass().equals(superClass))) {
			try {
				Method m = this.getClass()
						.getMethod(methodName, parameterTypes);

				if (m.getDeclaringClass().equals(superClass)) {
					throw new RejectedExecutionException("The method " + m
							+ " must be overwritten by class "
							+ this.getClass().getName());
				}
			} catch (NoSuchMethodException e) {
				throw new RejectedExecutionException("The method " + methodName
						+ "() must be overwritten by class "
						+ this.getClass().getName());
			}
		}
	}

	public void setClassLabel(String classLabel) {
		this.classLabel = classLabel;
	}
}
