/*
 * This file is part of Bielefeld Type Library (BTL).
 * 
 * Copyright(c) Frederic Siepmann http://opensource.cit-ec.de/projects/btl
 * 
 * This file may be licensed under the terms of of the GNU Lesser General Public
 * License Version 3 (the ``LGPL''), or (at your option) any later version.
 * 
 * Software distributed under the License is distributed on an ``AS IS'' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the LGPL for the
 * specific language governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this program. If not,
 * go to http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 * USA.
 * 
 * The development of this software was supported by the Excellence Cluster EXC
 * 277 Cognitive Interaction Technology. The Excellence Cluster EXC 277 is a
 * grant of the Deutsche Forschungsgemeinschaft (DFG) in the context of the
 * German Excellence Initiative.
 * 
 * Contributors: Leon Ziegler, Johannes Wienke, Matthias Schröder, Norman
 * Köster, Lukas Twardon, Torben Töniges, Annika Rothert, Marian Pohling, Denis
 * Schulze, Viktor Richter, Birte Carlmeyer, Maikel Linke
 */
package de.unibi.airobots.btl.data.asu;

import java.util.ArrayList;
import java.util.List;

/**
 * Represents a frame entry in frames.xml. Provides direct access to related
 * frames.
 * 
 * @author malinke
 */
public class FrameEntry {

	private final String name;

	private final List<FrameEntry> slots;

	private FrameEntry topFrame;

	/**
	 * Creates a new FrameEntry.
	 * 
	 * @param name
	 *            of this frame
	 */
	FrameEntry(final String name) {
		this.name = name;
		this.slots = new ArrayList<FrameEntry>();
	}

	/**
	 * Returns the name.
	 * 
	 * @return name of this frame
	 */
	public final String getName() {
		return this.name;
	}

	/**
	 * Returns the name of the top frame.
	 * 
	 * @return name of top frame
	 */
	public final String getTopName() {
		return this.topFrame.getName();
	}

	/**
	 * Creates a new list of all slot frames.
	 * 
	 * @return all slot frames
	 */
	public final List<FrameEntry> slotList() {
		List<FrameEntry> frames = new ArrayList<FrameEntry>(slots.size());
		frames.addAll(slots);
		return frames;
	}

	final void addSlot(final FrameEntry frame) {
		this.slots.add(frame);
	}

	final void setTopFrame(final FrameEntry frame) {
		this.topFrame = frame;
	}

	/**
	 * Compares this and all top frames (generalizations) with generalType.
	 * 
	 * @param generalType
	 *            to compare with
	 * @return true, if this or a super frame of this equals generalType
	 */
	public final boolean isOfType(final FrameEntry generalType) {
		if (this.equals(generalType)) {
			return true;
		}
		if (this.topFrame == null) {
			return false;
		}
		if (this.topFrame == this) {
			// avoid loops
			return false;
		}
		return this.topFrame.isOfType(generalType);
	}
}
